# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: config.proto

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='config.proto',
  package='',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=b'\n\x0c\x63onfig.proto\"\x8e\x03\n\x0b\x43onfigProto\x12\x11\n\x04seed\x18\x01 \x01(\x05:\x03\x37\x37\x37\x12\x16\n\x0b\x63_log_level\x18\x03 \x01(\x05:\x01\x32\x12\x17\n\x0cpy_log_level\x18\x04 \x01(\x05:\x01\x32\x12\x11\n\x05max_m\x18\x05 \x01(\x05:\x02\x31\x32\x12\x12\n\x06max_m0\x18\x06 \x01(\x05:\x02\x32\x34\x12\x1c\n\x0f\x65\x66_construction\x18\x07 \x01(\x05:\x03\x31\x35\x30\x12\x12\n\nlevel_mult\x18\t \x01(\x01\x12\x14\n\x0csave_remains\x18\n \x01(\x08\x12\x19\n\rhyper_threads\x18\x0b \x01(\x01:\x02\x31\x30\x12\x15\n\tblock_dim\x18\x0c \x01(\x05:\x02\x33\x32\x12\x16\n\tdist_type\x18\r \x01(\t:\x03\x64ot\x12\x1a\n\x12visited_table_size\x18\x11 \x01(\x05\x12\x1f\n\x11visited_list_size\x18\x0e \x01(\x05:\x04\x38\x31\x39\x32\x12\x11\n\x03nrz\x18\x0f \x01(\x08:\x04true\x12\x14\n\x0creverse_cand\x18\x10 \x01(\x08\x12\x1c\n\x0eheuristic_coef\x18\x12 \x01(\x01:\x04\x30.25'
)




_CONFIGPROTO = _descriptor.Descriptor(
  name='ConfigProto',
  full_name='ConfigProto',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='seed', full_name='ConfigProto.seed', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=777,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='c_log_level', full_name='ConfigProto.c_log_level', index=1,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='py_log_level', full_name='ConfigProto.py_log_level', index=2,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=2,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_m', full_name='ConfigProto.max_m', index=3,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=12,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_m0', full_name='ConfigProto.max_m0', index=4,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=24,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='ef_construction', full_name='ConfigProto.ef_construction', index=5,
      number=7, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=150,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='level_mult', full_name='ConfigProto.level_mult', index=6,
      number=9, type=1, cpp_type=5, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='save_remains', full_name='ConfigProto.save_remains', index=7,
      number=10, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='hyper_threads', full_name='ConfigProto.hyper_threads', index=8,
      number=11, type=1, cpp_type=5, label=1,
      has_default_value=True, default_value=float(10),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='block_dim', full_name='ConfigProto.block_dim', index=9,
      number=12, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=32,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='dist_type', full_name='ConfigProto.dist_type', index=10,
      number=13, type=9, cpp_type=9, label=1,
      has_default_value=True, default_value=b"dot".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='visited_table_size', full_name='ConfigProto.visited_table_size', index=11,
      number=17, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='visited_list_size', full_name='ConfigProto.visited_list_size', index=12,
      number=14, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=8192,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='nrz', full_name='ConfigProto.nrz', index=13,
      number=15, type=8, cpp_type=7, label=1,
      has_default_value=True, default_value=True,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='reverse_cand', full_name='ConfigProto.reverse_cand', index=14,
      number=16, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='heuristic_coef', full_name='ConfigProto.heuristic_coef', index=15,
      number=18, type=1, cpp_type=5, label=1,
      has_default_value=True, default_value=float(0.25),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=17,
  serialized_end=415,
)

DESCRIPTOR.message_types_by_name['ConfigProto'] = _CONFIGPROTO
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

ConfigProto = _reflection.GeneratedProtocolMessageType('ConfigProto', (_message.Message,), {
  'DESCRIPTOR' : _CONFIGPROTO,
  '__module__' : 'config_pb2'
  # @@protoc_insertion_point(class_scope:ConfigProto)
  })
_sym_db.RegisterMessage(ConfigProto)


# @@protoc_insertion_point(module_scope)
